# Copyright (c) 2024 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

import os
import pickle
import pytest

import numpy as np

import gen_tso.pandeia_io as jwst
from gen_tso.utils import ROOT

os.chdir(ROOT+'../tests')
# See tests/make_mocks.py for mocked data setup.


def test_saturation_level_perform_calculation_single():
    with open('mocks/perform_calculation_nircam_ssgrism.pkl', 'rb') as f:
        result = pickle.load(f)

    pixel_rate, full_well = jwst.saturation_level(result)
    expected_rate = 1396.5528564453125
    expected_well = 58100.00422843957
    np.testing.assert_allclose(pixel_rate, expected_rate)
    np.testing.assert_allclose(full_well, expected_well)


def test_saturation_level_perform_calculation_multi():
    with open('mocks/perform_calculation_miri_mrs_ts.pkl', 'rb') as f:
        result = pickle.load(f)

    pixel_rate, full_well = jwst.saturation_level(result)
    expected_rate = [
        204.37779236, 109.17165375,  32.75593948,   4.55676889,
    ]
    expected_well = [
        193654.99564498, 193655.00540975, 193655.004117  , 193654.99798687,
    ]
    np.testing.assert_allclose(pixel_rate, expected_rate)
    np.testing.assert_allclose(full_well, expected_well)


def test_saturation_level_tso_calculation_single():
    with open('mocks/tso_calculation_nircam_ssgrism.pkl', 'rb') as f:
        tso = pickle.load(f)

    pixel_rate, full_well = jwst.saturation_level(tso)
    expected_rate = [1354.6842041 , 1396.55285645]
    expected_well = [58100.00016363, 58100.00129109]
    np.testing.assert_allclose(pixel_rate, expected_rate)
    np.testing.assert_allclose(full_well, expected_well)


def test_saturation_level_tso_calculation_multi():
    with open('mocks/tso_calculation_miri_mrs_ts.pkl', 'rb') as f:
        tso = pickle.load(f)

    pixel_rate, full_well = jwst.saturation_level(tso)
    expected_rate = [
        198.24740601, 105.90003204,  31.80119514,   4.46783495,
        204.37779236, 109.17165375,  32.75593948,   4.55676889,
    ]
    expected_well = [
        193654.99147199, 193654.99886718, 193654.98493823, 193654.99820043,
        193654.99564498, 193655.00540975, 193655.004117  , 193654.99798687,
    ]
    np.testing.assert_allclose(pixel_rate, expected_rate)
    np.testing.assert_allclose(full_well, expected_well)


@pytest.mark.skip(reason='TBD')
def test_saturation_level_tso_calculation_get_max():
    pass


def test_exposure_time_nircam():
    inst = 'nircam'
    subarray = 'subgrism64'
    readout = 'rapid'
    ngroup = 90
    nint = 1
    exp_time = jwst.exposure_time(inst, subarray, readout, ngroup, nint)
    np.testing.assert_allclose(exp_time, 31.00063)


def test_exposure_time_miri():
    inst = 'miri'
    subarray = 'slitlessprism'
    readout = 'fastr1'
    ngroup = 30
    nint = 1
    exp_time = jwst.exposure_time(inst, subarray, readout, ngroup, nint)
    np.testing.assert_allclose(exp_time, 4.7712)


def test_exposure_time_bad_subarray():
    inst = 'nircam'
    subarray = 'nope'
    readout = 'rapid'
    ngroup = 90
    nint = 1
    exp_time = jwst.exposure_time(inst, subarray, readout, ngroup, nint)
    assert exp_time == 0.0


def test_exposure_time_bad_readout():
    inst = 'nircam'
    subarray = 'subgrism64'
    readout = 'nope'
    ngroup = 90
    nint = 1
    exp_time = jwst.exposure_time(inst, subarray, readout, ngroup, nint)
    assert exp_time == 0.0


def test_bin_search_exposure_time_nircam():
    inst = 'nircam'
    subarray = 'subgrism64'
    readout = 'rapid'
    ngroup = 90
    obs_time = 6.0
    nint, exp_time = jwst.bin_search_exposure_time(
        inst, subarray, readout, ngroup, obs_time,
    )
    assert nint == 697
    np.testing.assert_allclose(exp_time, 21607.43911)


def test_bin_search_exposure_time_miri():
    inst = 'miri'
    subarray = 'slitlessprism'
    readout = 'fastr1'
    ngroup = 30
    obs_time = 6.0
    nint, exp_time = jwst.bin_search_exposure_time(
        inst, subarray, readout, ngroup, obs_time,
    )
    assert nint == 4382
    np.testing.assert_allclose(exp_time, 21604.15264)


def test_bin_search_exposure_time_bad_subarray():
    inst = 'nircam'
    subarray = 'nope'
    readout = 'rapid'
    ngroup = 90
    obs_time = 6.0
    nint, exp_time = jwst.bin_search_exposure_time(
        inst, subarray, readout, ngroup, obs_time,
    )
    assert exp_time == 0.0


def test_bin_search_exposure_time_bad_readout():
    inst = 'nircam'
    subarray = 'subgrism64'
    readout = 'nope'
    ngroup = 90
    obs_time = 6.0
    nint, exp_time = jwst.bin_search_exposure_time(
        inst, subarray, readout, ngroup, obs_time,
    )
    assert exp_time == 0.0


def test__print_pandeia_exposure_config():
    with open('mocks/perform_calculation_nircam_ssgrism.pkl', 'rb') as f:
        result = pickle.load(f)
    config = result['input']['configuration']
    text = jwst._print_pandeia_exposure(config=config)
    assert text == 'Exposure time: 701.41 s (0.19 h)'


def test__print_pandeia_exposure_vals():
    inst = 'nircam'
    subarray = 'subgrism64'
    readout = 'rapid'
    ngroup = 90
    nint = 150
    text = jwst._print_pandeia_exposure(inst, subarray, readout, ngroup, nint)
    assert text == 'Exposure time: 4650.09 s (1.29 h)'


@pytest.mark.parametrize('format', [None, 'html', 'rich'])
def test__print_pandeia_exposure_format(format):
    with open('mocks/perform_calculation_nircam_ssgrism.pkl', 'rb') as f:
        result = pickle.load(f)
    config = result['input']['configuration']
    text = jwst._print_pandeia_exposure(config=config, format=format)
    assert text == 'Exposure time: 701.41 s (0.19 h)'


def test__print_pandeia_saturation_perform_calc():
    with open('mocks/perform_calculation_nircam_ssgrism.pkl', 'rb') as f:
        result = pickle.load(f)

    text = jwst._print_pandeia_saturation(reports=[result], format=None)
    expected_text = """Max fraction of saturation: 1.6%
ngroup below 80% saturation: 97
ngroup below 100% saturation: 122"""
    assert text == expected_text


def test__print_pandeia_saturation_tso_calc():
    with open('mocks/tso_calculation_miri_mrs_ts.pkl', 'rb') as f:
        tso = pickle.load(f)

    text = jwst._print_pandeia_saturation(reports=tso, format=None)
    expected_text = """Max fraction of saturation: 73.2%
ngroup below 80% saturation: 273
ngroup below 100% saturation: 341"""
    assert text == expected_text


def test__print_pandeia_saturation_values():
    pixel_rate, full_well = 1396.5528564453125, 58100.002280515175
    inst = 'nircam'
    subarray = 'subgrism64'
    readout = 'rapid'
    ngroup = 90
    text = jwst._print_pandeia_saturation(
        inst, subarray, readout, ngroup, pixel_rate, full_well,
        format=None,
    )
    expected_text = """Max fraction of saturation: 73.7%
ngroup below 80% saturation: 97
ngroup below 100% saturation: 122"""
    assert text == expected_text


@pytest.mark.skip(reason='TBD')
def test__print_pandeia_stats():
    pass


def test__print_pandeia_report_perform_calculation_single():
    with open('mocks/perform_calculation_nircam_ssgrism.pkl', 'rb') as f:
        result = pickle.load(f)

    report = jwst._print_pandeia_report([result], format=None)
    expected_report = """Exposure time: 701.41 s (0.19 h)
Max fraction of saturation: 1.6%
ngroup below 80% saturation: 97
ngroup below 100% saturation: 122

Signal-to-noise ratio        410.9
Extracted flux              2106.2  e-/s
Flux standard deviation        5.1  e-/s
Brightest pixel rate        1396.6  e-/s

Integrations:                         683
Duty cycle:                          0.33
Total exposure time:                701.4  s
First--last dt per exposure:        701.4  s
Reset--last dt per integration:     232.6  s

Reference wavelength:                    4.36  microns
Area of extraction aperture:             4.76  pixels
Area of background measurement:           6.3  pixels
Background surface brightness:            0.3  MJy/sr
Total sky flux in background aperture:   4.45  e-/s
Total flux in background aperture:      65.97  e-/s
Background flux fraction from scene:     0.93
Number of cosmic rays:      0.0002  events/pixel/read"""
    assert report == expected_report



def test__print_pandeia_report_perform_calculation_multi():
    with open('mocks/perform_calculation_miri_mrs_ts.pkl', 'rb') as f:
        result = pickle.load(f)

    report = jwst._print_pandeia_report(result, format=None)
    expected_report = """Exposure time: 27858.63 s (7.74 h)
Max fraction of saturation: 73.2%
ngroup below 80% saturation: 273
ngroup below 100% saturation: 341

Signal-to-noise ratio       1334.2    1265.2     960.1     179.2
Extracted flux               225.1     147.4      46.2       2.0  e-/s
Flux standard deviation        0.2       0.1       0.0       0.0  e-/s
Brightest pixel rate         204.4     109.2      32.8       4.6  e-/s

Integrations:                          40
Duty cycle:                          0.99
Total exposure time:              27858.6  s
First--last dt per exposure:      27858.6  s
Reset--last dt per integration:   27639.4  s

Reference wavelength:                    5.35   8.15  12.50  19.29  microns
Area of extraction aperture:             8.20   5.21   2.98   1.61  pixels
Area of background measurement:          79.4   50.5   28.9   15.6  pixels
Background surface brightness:            1.0   10.9   43.0  165.5  MJy/sr
Total sky flux in background aperture:   0.29   0.82   2.98   2.44  e-/s
Total flux in background aperture:       0.85   1.47   3.19   2.54  e-/s
Background flux fraction from scene:     0.66   0.45   0.06   0.04
Number of cosmic rays:      0.3122  events/pixel/read"""
    assert report == expected_report


def test__print_pandeia_report_tso_calculation_single():
    with open('mocks/tso_calculation_nircam_ssgrism.pkl', 'rb') as f:
        result = pickle.load(f)

    report = jwst._print_pandeia_report(result, format=None)
    expected_report = """Exposure time: 21545.44 s (5.98 h)
Max fraction of saturation: 73.7%
ngroup below 80% saturation: 97
ngroup below 100% saturation: 122

Signal-to-noise ratio       3484.8
Extracted flux              2043.0  e-/s
Flux standard deviation        0.6  e-/s
Brightest pixel rate        1354.7  e-/s

                               in-transit  out-transit
Integrations:                         243      452
Duty cycle:                          0.98     0.98
Total exposure time:               7533.2  14012.3  s
First--last dt per exposure:       7533.2  14012.3  s
Reset--last dt per integration:    7366.4  13702.1  s

Reference wavelength:                    4.36  microns
Area of extraction aperture:             4.76  pixels
Area of background measurement:           6.3  pixels
Background surface brightness:            0.3  MJy/sr
Total sky flux in background aperture:   4.45  e-/s
Total flux in background aperture:      64.12  e-/s
Background flux fraction from scene:     0.93
Number of cosmic rays:      0.0072  events/pixel/read"""
    assert report == expected_report


def test__print_pandeia_report_tso_calculation_multi():
    with open('mocks/tso_calculation_miri_mrs_ts.pkl', 'rb') as f:
        tso = pickle.load(f)

    report = jwst._print_pandeia_report(tso, format=None)
    expected_report = """Exposure time: 20893.28 s (5.80 h)
Max fraction of saturation: 73.2%
ngroup below 80% saturation: 273
ngroup below 100% saturation: 341

Signal-to-noise ratio        657.0     623.0     472.3      87.5
Extracted flux               218.4     142.9      44.8       1.9  e-/s
Flux standard deviation        0.3       0.2       0.1       0.0  e-/s
Brightest pixel rate         198.2     105.9      31.8       4.5  e-/s

                               in-transit  out-transit
Integrations:                          10       20
Duty cycle:                          0.99     0.99
Total exposure time:               6962.6  13927.9  s
First--last dt per exposure:       6962.6  13927.9  s
Reset--last dt per integration:    6909.8  13819.7  s

Reference wavelength:                    5.35   8.15  12.50  19.29  microns
Area of extraction aperture:             8.20   5.21   2.98   1.61  pixels
Area of background measurement:          79.4   50.5   28.9   15.6  pixels
Background surface brightness:            1.0   10.9   43.0  165.5  MJy/sr
Total sky flux in background aperture:   0.29   0.82   2.98   2.44  e-/s
Total flux in background aperture:       0.83   1.45   3.18   2.54  e-/s
Background flux fraction from scene:     0.65   0.44   0.06   0.04
Number of cosmic rays:      0.3122  events/pixel/read"""
    assert report == expected_report


def test_tso_print_plain(capsys):
    with open('mocks/tso_calculation_nircam_ssgrism.pkl', 'rb') as f:
        result = pickle.load(f)
    # Make some troublesome figures
    result['report_in']['scalar']['duty_cycle'] = 0.25
    result['report_in']['input']['configuration']['detector']['ngroup'] = 130
    result['report_out']['scalar']['duty_cycle'] = 0.25
    result['report_out']['input']['configuration']['detector']['ngroup'] = 130

    jwst.tso_print(result, format=None)
    captured = capsys.readouterr()
    expected_report = (
        "Exposure time: 31014.40 s (8.62 h)\r\n"
        "Max fraction of saturation: 106.4%\r\n"
        "ngroup below 80% saturation: 97\r\n"
        "ngroup below 100% saturation: 122\r\n"
        "\r\n"
        "Signal-to-noise ratio       3484.8\r\n"
        "Extracted flux              2043.0  e-/s\r\n"
        "Flux standard deviation        0.6  e-/s\r\n"
        "Brightest pixel rate        1354.7  e-/s\r\n"
        "\r\n"
        "                               in-transit  out-transit\r\n"
        "Integrations:                         243      452\r\n"
        "Duty cycle:                          0.25     0.25\r\n"
        "Total exposure time:               7533.2  14012.3  s\r\n"
        "First--last dt per exposure:       7533.2  14012.3  s\r\n"
        "Reset--last dt per integration:    7366.4  13702.1  s\r\n"
        "\r\n"
        "Reference wavelength:                    4.36  microns\r\n"
        "Area of extraction aperture:             4.76  pixels\r\n"
        "Area of background measurement:           6.3  pixels\r\n"
        "Background surface brightness:            0.3  MJy/sr\r\n"
        "Total sky flux in background aperture:   4.45  e-/s\r\n"
        "Total flux in background aperture:      64.12  e-/s\r\n"
        "Background flux fraction from scene:     0.93\r\n"
        "Number of cosmic rays:      0.0072  events/pixel/read\r\n"
    )
    assert captured.out == expected_report


@pytest.mark.skip(reason='Having some issues with capsys IO')
def test_tso_print_rich(capsys):
    with open('mocks/tso_calculation_nircam_ssgrism.pkl', 'rb') as f:
        result = pickle.load(f)
    # Make some troublesome figures
    result['report_in']['scalar']['duty_cycle'] = 0.25
    result['report_in']['input']['configuration']['detector']['ngroup'] = 130
    result['report_out']['scalar']['duty_cycle'] = 0.25
    result['report_out']['input']['configuration']['detector']['ngroup'] = 130

    jwst.tso_print(result)
    captured = capsys.readouterr()
    expected_report = (
        "Exposure time: 31014.40 s (8.62 h)\r\n"
        "Max fraction of saturation: 106.4%\r\n"
        "ngroup below 80% saturation: 97\r\n"
        "ngroup below 100% saturation: 122\r\n"
        "\r\n"
        "Signal-to-noise ratio       3484.8\r\n"
        "Extracted flux              2043.0  e-/s\r\n"
        "Flux standard deviation        0.6  e-/s\r\n"
        "Brightest pixel rate        1354.7  e-/s\r\n"
        "\r\n"
        "                               in-transit  out-transit\r\n"
        "Integrations:                         243      452\r\n"
        "Duty cycle:                          0.25     0.25\r\n"
        "Total exposure time:               7533.2  14012.3  s\r\n"
        "First--last dt per exposure:       7533.2  14012.3  s\r\n"
        "Reset--last dt per integration:    7366.4  13702.1  s\r\n"
        "\r\n"
        "Reference wavelength:                    4.36  microns\r\n"
        "Area of extraction aperture:             4.76  pixels\r\n"
        "Area of background measurement:           6.3  pixels\r\n"
        "Background surface brightness:            0.3  MJy/sr\r\n"
        "Total sky flux in background aperture:   4.45  e-/s\r\n"
        "Total flux in background aperture:      64.12  e-/s\r\n"
        "Background flux fraction from scene:     0.93\r\n"
        "Number of cosmic rays:      0.0072  events/pixel/read\r\n"
    )
    assert captured.out == expected_report




# Copyright (c) 2021-2024 Patricio Cubillos
# Pyrat Bay is open-source software under the GPL-2.0 license
# Copyright (c) 2024 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

"""
Some of these routines have been lifted from the pyratbay package
When pyratbay 2.0 gets release, remove these and add dependency.
"""

__all__ = [
    'ROOT',
    'Tophat',
    'constant_resolution_spectrum',
    'bin_spectrum',
    'read_spectrum_file',
    'collect_spectra',
    'format_text',
    'pretty_print_target',
]

import operator
import os

import numpy as np
from prompt_toolkit.formatted_text import FormattedText
from pyratbay.tools import u
from pyratbay.spectrum import PassBand
from shiny import ui

ROOT = os.path.realpath(os.path.dirname(__file__)) + '/'
from .catalogs.utils import as_str


class Tophat(PassBand):
    """
    A Filter passband object with a tophat-shaped passband.
    """
    def __init__(
            self, wl0, half_width,
            name='tophat', ignore_gaps=False,
        ):
        """
        Parameters
        ----------
        wl0: Float
            The passband's central wavelength (um units).
        half_width: Float
            The passband's half-width (um units).
        name: Str
            A user-defined name for the filter when calling str(self),
            e.g., to identify the instrument provenance of this filter.

        Examples
        --------
        >>> import pyratbay.spectrum as ps
        >>> import matplotlib.pyplot as plt
        >>> import numpy as np

        >>> hat = ps.Tophat(4.5, 0.5)

        >>> # Evaluate over a wavelength array (um units):
        >>> wl = np.arange(3.5, 5.5, 0.001)
        >>> out_wl, out_response = hat(wl)

        >>> plt.figure(1)
        >>> plt.clf()
        >>> plt.plot(out_wl, out_response)
        >>> plt.plot(hat.wl, hat.response)  # Same variables
        >>> # Note wl differs from hat.wl, but original array can be used as:
        >>> plt.plot(wl[hat.idx], hat.response)

        >>> # Evaluate over a wavenumber array:
        >>> wn = 1e4 / wl
        >>> hat(wn=wn)
        >>> plt.figure(1)
        >>> plt.clf()
        >>> plt.plot(hat.wn, hat.response, dashes=(5,3))
        >>> plt.plot(wn[hat.idx], out_response)
        """
        self.wl0 = wl0
        self.half_width = half_width
        # Read filter wavenumber and transmission curves:
        self.wn0 = 1.0 / (self.wl0 * u('um'))
        self.ignore_gaps = ignore_gaps

        self.name = name

    def __repr__(self):
        return f'pyratbay.spectrum.Tophat({self.wl0}, {self.half_width})'

    def __str__(self):
        return f'{self.name}_{self.wl0}um'

    def __call__(self, wl=None, wn=None):
        """
        Interpolate filter response function at specified spectral array.
        The response funciton is normalized such that the integral over
        wavenumber equals one.

        Parameters
        ----------
        wl: 1D float array
            Wavelength array at which evaluate the passband response's
            in micron units.
            (only one of wl or wn should be provided on call)
        wn: 1D float array
            Wavenumber array at which evaluate the passband response's
            in cm-1 units.
            (only one of wl or wn should be provided on call)
        ignore_gaps: Bool
            If True, set variables to None when there are no points
            inside the band. A ps.bin_spectrum() call on such band
            will return np.nan.
            Otherwise the code will throw an IndexError when attempting to
            index from an empty array.

        Defines
        -------
        self.response  Normalized interpolated response function
        self.idx       Wavenumber indices
        self.wn        Passband's wavenumber array
        self.wl        Passband's wavelength array

        Returns
        -------
        out_wave: 1D float array
            Same as self.wl or self.wn depending on the input argument.
        out_response: 1D float array
            Same as self.response

        Examples
        --------
        >>> # See examples in help(ps.Tophat.__init__)
        """
        if not operator.xor(wl is None, wn is None):
            raise ValueError(
                'Either provide wavelength or wavenumber array, not both'
            )
        input_is_wl = wn is None
        if input_is_wl:
            wn = 1.0 / (wl*u('um'))

        sign = np.sign(np.ediff1d(wn))
        if not (np.all(sign == 1) or np.all(sign == -1)):
            raise ValueError(
                'Input wavelength/wavenumber array must be strictly '
                'increasing or decreasing'
            )
        sign = sign[0]
        nwave = len(wn)

        wl_low  = self.wl0 - self.half_width
        wl_high = self.wl0 + self.half_width

        wn_low = 1.0 / (wl_high*u('um'))
        wn_high = 1.0 / (wl_low*u('um'))
        idx = (wn >= wn_low) & (wn <= wn_high)
        indices = np.where(idx)[0]

        if self.ignore_gaps and len(indices) == 0:
            self.idx = None
            self.response = None
            self.wn = None
            self.wl = None
            return None, None

        # One spectral point as margin:
        idx_first = indices[0]
        idx_last = indices[-1] + 1

        if idx_first > 0:
            idx_first -= 1
        if idx_last < nwave:
            idx_last += 1

        if sign < 0.0:
            self.idx = np.flip(np.arange(idx_first, idx_last))
        else:
            self.idx = np.arange(idx_first, idx_last)

        self.wn = wn[self.idx]
        self.response = np.array(idx[self.idx], np.double)
        self.response /= np.trapz(self.response, self.wn)

        self.wl = 1.0 / (self.wn * u('um'))
        if input_is_wl:
            out_wave = self.wl
        else:
            out_wave = self.wn

        return out_wave, self.response


def constant_resolution_spectrum(wave_min, wave_max, resolution):
    """
    Compute a constant resolving-power sampling array.

    Parameters
    ----------
    wave_min: Float
        Lower spectral boundary.  This could be either a wavelength
        or a wavenumber. This is agnositc of units.
    wave_max: Float
        Upper spectral boundary.  This could be either a wavelength
        or a wavenumber. This is agnositc of units.
    resolution: Float
        The sampling resolving power: R = wave / delta_wave.

    Returns
    -------
    wave: 1D float array
        A spectrum array with the given resolving power.

    Examples
    --------
    >>> import numpy as np
    >>> import pyratbay.spectrum as ps

    >>> # A low-resolution wavelength sampling:
    >>> wl_min = 0.5
    >>> wl_max = 4.0
    >>> resolution = 5.5
    >>> wl = ps.constant_resolution_spectrum(wl_min, wl_max, resolution)
    >>> print(wl)
    [0.5        0.6        0.72       0.864      1.0368     1.24416
     1.492992   1.7915904  2.14990848 2.57989018 3.09586821 3.71504185]
    >>> # The actual resolution matches the input:
    >>> wl_mean = 0.5*(wl[1:]+wl[:-1])
    >>> print(wl_mean/np.ediff1d(wl))
    [5.5 5.5 5.5 5.5 5.5 5.5 5.5 5.5 5.5 5.5 5.5]
    """
    f = 0.5 / resolution
    g = (1.0+f) / (1.0-f)

    nwave = int(np.ceil(-np.log(wave_min/wave_max) / np.log(g)))
    wave = wave_min * g**np.arange(nwave)
    return wave


def bin_spectrum(bin_wl, wl, spectrum, half_widths=None, ignore_gaps=False):
    """
    Bin down a spectrum.

    Parameters
    ----------
    bin_wl: 1D float array
        Central wavelength (um) of the desired binned down spectra.
    wl: 1D float array
        Wavelength samples of the original spectrum.
    spectrum: 1D float array
        Spectral values to be binned down.
    half_widths: 1D float array
        The bin half widths (um).
        If None, assume that the bin edges are at the mid-points
        of the bin_wl array.

    Returns
    -------
    bin_spectrum: 1D float array
        The binned spectrum.

    Notes
    -----
    Probably bad things will happen if bin_wl has a similar
    or coarser resolution than wl.

    Examples
    --------
    >>> import pyratbay.spectrum as ps
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt

    >>> # Make a noisy high-resolution signal
    >>> wl = ps.constant_resolution_spectrum(1.0, 3.0, resolution=5000)
    >>> spectrum = np.sin(3.14*wl) + np.random.normal(1.5, 0.1, len(wl))
    >>> # Bin it down:
    >>> bin_wl = ps.constant_resolution_spectrum(1.0, 3.0, resolution=125)
    >>> bin_spectrum = ps.bin_spectrum(bin_wl, wl, spectrum)

    >>> # Compare original and binned signals
    >>> plt.figure(0)
    >>> plt.clf()
    >>> plt.plot(wl, spectrum, '.', ms=2, color='gray')
    >>> plt.plot(bin_wl, bin_spectrum, color='red')
    """
    if half_widths is None:
        half_widths = np.ediff1d(bin_wl, 0, 0)
        half_widths[0] = half_widths[1]
        half_widths[-1] = half_widths[-2]
        half_widths /= 2.0
    bands = [
        Tophat(wl0, half_width, ignore_gaps=ignore_gaps)
        for wl0, half_width in zip(bin_wl, half_widths)
    ]
    nbands = len(bands)
    band_flux = np.zeros(nbands)
    for i,band in enumerate(bands):
        band_wl, response = band(wl)
        if band.idx is None:
            band_flux[i] = np.nan
        else:
            band_flux[i] = np.trapz(spectrum[band.idx]*response, band.wn)
    return band_flux


def read_spectrum_file(file, on_fail=None):
    """
    Parameters
    ----------
    file: String
        Spectrum file to read (transit depth, eclipse depth, or stellar SED)
        This is a plain-text file with two columns (white space separater)
        First column is the wavelength, second is the depth/flux.
        Should be readable by numpy.loadtxt().
    on_fail: String
        if 'warning' raise a warning.
        if 'error' raise an error.

    Examples
    --------
    >>> import gen_tso.utils as u

    >>> file = f'{u.ROOT}data/models/WASP80b_transit.dat'
    >>> spectra = u.read_spectrum_file(file, on_fail='warning')
    """
    try:
        data = np.loadtxt(file, unpack=True)
        wl, depth = data
    except ValueError as error:
        wl = None
        depth = None
        error_msg = (
            f'Error, could not load spectrum file: {repr(file)}\n{error}'
        )
        if on_fail == 'warning':
            print(error_msg)
        if on_fail == 'error':
            raise ValueError(error_msg)

    path, label = os.path.split(file)
    if label.endswith('.dat') or label.endswith('.txt'):
        label = label[0:-4]
    return label, wl, depth


def collect_spectra(folder, on_fail=None):
    """
    Parameters
    ----------
    on_fail: String
        if 'warning' raise a warning.
        if 'error' raise an error.

    Examples
    --------
    >>> import gen_tso.utils as u

    >>> folder = f'{u.ROOT}data/models/'
    >>> spectra = u.collect_spectra(folder, on_fail=None)
    """
    files = os.listdir(folder)
    transit_files = [
        file for file in sorted(files)
        if 'transit' in file or 'transmission' in file
    ]
    eclipse_files = [
        file for file in sorted(files)
        if 'eclipse' in file or 'emission' in file
    ]
    sed_files = [
        file for file in sorted(files)
        if 'sed' in file or 'star' in file
    ]

    transit_spectra = {}
    for file in transit_files:
        label, wl, depth = read_spectrum_file(f'{folder}/{file}', on_fail)
        if wl is not None:
            transit_spectra[label] = {'wl': wl, 'depth': depth}

    eclipse_spectra = {}
    for file in eclipse_files:
        label, wl, depth = read_spectrum_file(f'{folder}/{file}', on_fail)
        if wl is not None:
            eclipse_spectra[label] = {'wl': wl, 'depth': depth}

    sed_spectra = {}
    for file in sed_files:
        label, wl, model = read_spectrum_file(f'{folder}/{file}', on_fail)
        if wl is not None:
            sed_spectra[label] = {'wl': wl, 'flux': model}

    return transit_spectra, eclipse_spectra, sed_spectra


def format_text(text, warning=False, danger=False, format=None):
    """
    Return a colorful text depending on requested format and warning
    or danger flags.

    Parameters
    ----------
    text: String
        A text to print with optional richer format.
    warning: Bool
        If True, format as warning text (orange color).
    danger: Bool
        If True, format as danger text (red color).
        If True, overrides warning.
    format: String
        If None return plain text.
        If 'html' return HTML formatted text.
        If 'rich' return formatted text to be printed with prompt_toolkit.

    See also
    --------
    gen_tso.pandeia_io.tso_print
        
    Examples
    --------
    >>> import gen_tso.utils as u
    >>> text = 'WASP-80 b'
    >>> plain = u.format_text(text, danger=True)
    >>> normal = u.format_text(text, warning=False, danger=False, format='html')
    >>> html = u.format_text(text, danger=True, format='html')
    >>> rich = u.format_text(text, danger=True, format='rich')

    >>> warned = u.format_text(text, warning=True, format='html')
    >>> danger1 = u.format_text(text, danger=True, format='html')
    >>> danger2 = u.format_text(text, warning=True, danger=True, format='html')
    """
    status = 'normal'
    if danger:
        status = 'danger'
    elif warning:
        status = 'warning'

    if format is None or status=='normal':
        return text

    if format == 'html':
        text_value = f'<span class="{status}">{text}</span>'
    elif format == 'rich':
        text_value = f'<{status}>{text}</{status}>'
    return text_value


def pretty_print_target(target):
    """
    Print a target's info to HTML text.
    Must look pretty.
    """
    rplanet = as_str(target.rplanet, '.3f', '---')
    mplanet = as_str(target.mplanet, '.3f', '---')
    sma = as_str(target.sma, '.3f', '---')
    rprs = as_str(target.rprs, '.3f', '---')
    ars = as_str(target.ars, '.3f', '---')
    period = as_str(target.period, '.3f', '---')
    t_dur = as_str(target.transit_dur, '.3f', '---')
    eq_temp = as_str(target.eq_temp, '.1f', '---')

    rstar = as_str(target.rstar, '.3f', '---')
    mstar = as_str(target.mstar, '.3f', '---')
    logg = as_str(target.logg_star, '.2f', '---')
    metal = as_str(target.metal_star, '.2f', '---')
    teff = as_str(target.teff, '.1f', '---')
    ks_mag = as_str(target.ks_mag, '.2f', '---')

    status = 'confirmed' if target.is_confirmed else 'candidate'
    mplanet_label = 'M*sin(i)' if target.is_min_mass else 'mplanet'
    if len(target.aliases) > 0:
        aliases = f'aliases = {target.aliases}'
    else:
        aliases = ''

    planet_info = ui.HTML(
        f'planet = {target.planet} <br>'
        f'is_transiting = {target.is_transiting}<br>'
        f'status = {status} planet<br><br>'
        f"rplanet = {rplanet} r_earth<br>"
        f"{mplanet_label} = {mplanet} m_earth<br>"
        f"semi-major axis = {sma} AU<br>"
        f"period = {period} d<br>"
        f"equilibrium temp = {eq_temp} K<br>"
        f"transit_dur (T14) = {t_dur} h<br>"
        f"rplanet/rstar = {rprs}<br>"
        f"a/rstar = {ars}<br>"
    )

    star_info = ui.HTML(
        f'host = {target.host}<br>'
        f'is JWST host = {target.is_jwst}<br>'
        f'<br><br>'
        f"rstar = {rstar} r_sun<br>"
        f"mstar = {mstar} m_sun<br>"
        f"log_g = {logg}<br>"
        f"metallicity = {metal}<br>"
        f"effective temp = {teff} K<br>"
        f"Ks_mag = {ks_mag}<br>"
        f"RA = {target.ra:.3f} deg<br>"
        f"dec = {target.dec:.3f} deg<br>"
    )

    return planet_info, star_info, aliases






# Copyright (c) 2024 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)


# Pyrat-Bay version:
TSO_VER = 0  # Major version
TSO_MIN = 4  # Minor version
TSO_REV = '0'  # Revision


__version__ = f'{TSO_VER}.{TSO_MIN}.{TSO_REV}'




# Copyright (c) 2024 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

import gen_tso.pandeia_io as jwst
import numpy as np


def test_perform_calculation():
    pando = jwst.PandeiaCalculation('niriss', 'soss')
    # Set the stellar scene:
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.637)

    # Set a NIRSpec observation
    disperser = 'gr700xd'
    filter = 'clear'
    readout = 'nisrapid'
    subarray = 'substrip96'
    aperture = 'soss'
    ngroup = 3
    nint = 984

    report = pando.perform_calculation(
        ngroup, nint, disperser, filter, subarray, readout, aperture,
    )
    # assert


#def test_read_noise_variance():
#    report_config = {
#        'mode': '',
#        'aperture': '',
#    }
#    inputs = (
#        ('miri', 'mrs_ts', 'ch1'),
#        ('miri', 'lrsslitless', 'imager'),
#        ('nircam', 'ssgrism', 'lw'),
#        ('nirspec', 'bots',  's1600a1'),
#        ('niriss', 'soss', 'soss'),
#    )
#    for input in inputs:
#        inst, mode, aperture = input
#        ins_config = get_instrument_config('jwst', inst)
#        pando = jwst.PandeiaCalculation(inst, mode)
#        #print(pando.calc['configuration']['instrument']['aperture'])
#        info_dict = ins_config['detector_config']
#        report_config['mode'] = mode
#        report_config['aperture'] = aperture
#
#        #report_config = report['input']['configuration']['instrument']
#        if report_config['mode'] == 'mrs_ts':
#            aperture = report_config['aperture']
#            aperture = ins_config['aperture_config'][aperture]['detector']
#
#        if aperture not in ins_config['detector_config']:
#            aperture = 'default'
#
#        read_noise = ins_config['detector_config'][aperture]['rn']
#        if isinstance(read_noise, dict):
#            read_noise = read_noise['default']
#        print(inst, mode, read_noise)


def test_tso_calculation_single():
    pando = jwst.PandeiaCalculation('nirspec', 'bots')
    # Set the stellar scene:
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.637)
    depth_model = np.loadtxt('data/models/WASP80b_transit.dat', unpack=True)

    # Set a NIRSpec observation
    disperser = 'g395h'
    filter = 'f290lp'
    readout = 'nrsrapid'
    subarray = 'sub2048'
    ngroup = 16

    transit_dur = 2.753
    obs_dur = 7.1
    obs_type = 'transit'
    tso = pando.tso_calculation(
        obs_type, transit_dur, obs_dur, depth_model,
        ngroup, disperser, filter, subarray, readout,
    )
    # assert


def test_tso_calculation_multiple():
    pando = jwst.PandeiaCalculation('miri', 'mrs_ts')
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.637)
    depth_model = np.loadtxt('data/models/WASP80b_transit.dat', unpack=True)

    disperser = 'short'
    filter = None
    readout = 'fastr1'
    subarray = 'full'
    ngroup = 16
    aperture = ['ch1', 'ch2']

    transit_dur = 2.753
    obs_dur = 7.1
    obs_type = 'transit'
    tso = pando.tso_calculation(
        obs_type, transit_dur, obs_dur, depth_model,
        ngroup, disperser, filter, subarray, readout, aperture,
    )
    # assert


def test_calc_saturation_single():
    instrument = 'nircam'
    mode = 'ssgrism'
    pando = jwst.PandeiaCalculation(instrument, mode)
    pando.set_scene('phoenix', 'k2v', '2mass,ks', 8.351)
    pixel_rate, full_well = pando.get_saturation_values(
        disperser='grismr', filter='f444w',
        readout='rapid', subarray='subgrism64',
    )
    np.testing.assert_almost_equal(pixel_rate, 1335.1112060546875)
    np.testing.assert_almost_equal(full_well, 58100.00314689601)


def test_calc_saturation_multiple():
    inst = 'miri'
    mode = 'mrs_ts'
    pando = jwst.PandeiaCalculation(inst, mode)    
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)
    ngroup = 2
    disperser = 'short'
    filter = None
    subarray = 'full'
    readout = 'fastr1'
    aperture = 'ch1 ch2 ch3 ch4'.split()

    pixel_rate, full_well = pando.get_saturation_values(
        disperser, filter, subarray, readout, ngroup, aperture,
    )
    expected_rate = [204.3777924, 109.1716537, 32.7559395, 4.5567689]
    expected_well = [
        193655.0135037985,
        193655.00840719877,
        193655.00325247002,
        193655.00765403986,
    ]
    np.testing.assert_almost_equal(pixel_rate, expected_rate)
    np.testing.assert_almost_equal(full_well, expected_well)





import pickle
import matplotlib.pyplot as plt
import numpy as np
import scipy.interpolate as sip
from astropy.io import fits

from pandeia.engine.calc_utils import (
    build_default_calc,
)
from pandeia.engine.instrument_factory import InstrumentFactory
from pandeia.engine.perform_calculation import perform_calculation

from gen_tso.pandeia_io import get_configs
import gen_tso.pandeia_io as jwst


def get_throughputs(calc, filter_name):
    wl = np.arange(0.5, 15.001, 0.0002)
    qe = {}
    calculation = calc.calc
    if calc.mode == 'bots':
        disperser, filter = filter_name.split('/')
        calculation['configuration']['instrument']['disperser'] = disperser
    else:
        filter = filter_name
    calculation['configuration']['instrument']['filter'] = filter
    inst = InstrumentFactory(
        config=calculation['configuration'], webapp=True,
    )
    qe[filter_name] = inst.get_total_eff(wl)
    return wl, qe


def mask_througput(wl_arr, qe, wl_response, response, thin=50, threshold=1e-3):
    """
    i = 6
    norm = np.amax(fluxes[i]) / np.amax(qe[filters[i]])
    wl_mask, qe_mask = masked_througput(wl_arr, qe[filters[i]], wl[i], fluxes[i]/norm)
    """
    # resample fluxes into wl_arr
    interp_resp = sip.interp1d(
        wl_response, response, bounds_error=False, fill_value=0.0,
    )
    # mask fluxes
    resp_mask = interp_resp(wl_arr) > threshold
    masked_response = qe * resp_mask

    # add margin of 1 before and after the mask edges
    edges = np.where(np.ediff1d(resp_mask*1))[0]

    # decimate
    mask = np.zeros(len(wl_arr), bool)
    mask[::thin] = True
    mask &= resp_mask
    mask[edges] = True
    mask[edges+1] = True
    return wl_arr[mask], masked_response[mask]


# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

def acquisition_throughputs():
    #insts = get_configs(obs_type='spectroscopy')
    insts = get_configs(obs_type='acquisition')

    inst = insts[3]
    instrument = inst['instrument'].lower()
    mode = inst['mode']
    telescope = 'jwst'
    calculation = build_default_calc(telescope, instrument, mode)

    throughputs = {}
    qe = {}
    #if instrument == 'niriss':
    #    filters = ['imager', 'nrm']
    #    key = 'aperture'
    #else:
    filters = list(inst['filters'])
    key = 'filter'
    if instrument == 'miri':
        wl = np.arange(4.5, 28.001, 0.0002)
    else:
        wl = np.arange(0.5, 15.001, 0.0002)
    for filter in filters:
        calculation['configuration']['instrument'][key] = filter
        ins = InstrumentFactory(
            config=calculation['configuration'], webapp=True,
        )
        qe[filter] = ins.get_total_eff(wl)

        threshold = 1e-5 if filter=='fnd' else 1e-3
        wl_mask, qe_mask = mask_througput(
            wl, qe[filter], wl, qe[filter], thin=50, threshold=threshold,
        )
        throughputs[filter] = {'wl': wl_mask, 'response': qe_mask}

    all_throughputs = {}
    for subarray in list(inst['subarrays']):
        all_throughputs[subarray] = throughputs
    t_file = f'../data/throughputs/throughputs_{instrument}_{mode}.pickle'
    with open(t_file, 'wb') as handle:
        pickle.dump(all_throughputs, handle, protocol=pickle.HIGHEST_PROTOCOL)

    plt.figure(0)
    plt.clf()
    for filter, throughput in throughputs.items():
        plt.plot(throughput['wl'], throughput['response'], label=filter)
    plt.ylim(0, 1)
    plt.legend(loc='best')



def generate_all_nirspec_throughputs():
    insts = get_configs(obs_type='spectroscopy')

    inst = insts[2]
    instrument = inst['instrument'].lower()
    mode = inst['mode']
    telescope = 'jwst'
    calculation = build_default_calc(telescope, instrument, mode)

    readout = 'nrsrapid'

    #calc = jwst.Calculation(instrument, mode)
    scene = jwst.make_scene(
        sed_type='flat', sed_model='flam',
        norm_band='2mass,ks', norm_magnitude=15.0,
    )
    filters = inst.get_configs('filters')
    subarrays = inst.get_configs('subarrays')

    nint = 1
    ngroup = 2

    throughputs = {}
    for subarray in subarrays:
        throughputs[subarray] = {}
        for grating in filters:
            disperser, filter = grating.split('/')
            print(subarray, disperser, filter)
            try:
                calc.perform_calculation(
                    nint, ngroup, filter, readout, subarray, disperser,
                )
            except:
                print(
                    f'Pandeia failed for: {instrument}, {mode}, '
                    f'{subarray}, {disperser}, {filter}'
                )
                continue
            wave, flux = calc.report['1d']['extracted_flux']
            wl_arr, qe = get_throughputs(calc, grating)

            norm = np.amax(flux) / np.amax(qe[grating])
            wl_mask, qe_mask = mask_througput(
                wl_arr, qe[grating], wave, flux/norm, thin=50,
            )
            throughputs[subarray][grating] = {'wl': wl_mask, 'response': qe_mask}

    t_file = f'../data/throughputs/throughputs_{instrument}_{mode}.pickle'
    with open(t_file, 'wb') as handle:
        pickle.dump(throughputs, handle, protocol=pickle.HIGHEST_PROTOCOL)


def generate_all_nircam_throughputs():
    insts = get_configs(obs_type='spectroscopy')

    inst = insts[1]
    instrument = inst['instrument'].lower()
    mode = inst['mode']
    telescope = 'jwst'
    calculation = build_default_calc(telescope, instrument, mode)

    #calc = jwst.Calculation(instrument, mode)
    scene = jwst.make_scene(
        sed_type='flat', sed_model='flam',
        norm_band='2mass,ks', norm_magnitude=15.0,
    )
    filters = list(inst['filters'])
    subarrays = list(inst['subarrays'])
    #readout = 'rapid'
    #disperser = 'grismr'
    disperser = calculation['configuration']['instrument']['disperser']
    readout = calculation['configuration']['detector']['readout_pattern']
    calculation['configuration']['detector']['ngroup'] = 2
    #nint = 1

    throughputs = {}
    for subarray in subarrays:
        throughputs[subarray] = {}
        for filter in filters:
            print(subarray, disperser, filter)
            try:
                calc.perform_calculation(
                    nint, ngroup, filter, readout, subarray, disperser,
                )
            except:
                print(
                    f'Pandeia failed for: {instrument}, {mode}, '
                    f'{subarray}, {disperser}, {filter}'
                )
                continue
            wave, flux = calc.report['1d']['extracted_flux']
            wl_arr, qe = get_throughputs(calc, [filter])

            norm = np.amax(flux) / np.amax(qe[filter])
            wl_mask, qe_mask = mask_througput(
                wl_arr, qe[filter], wave, flux/norm, thin=50,
            )
            throughputs[subarray][filter] = {'wl': wl_mask, 'response': qe_mask}

    t_file = f'../data/throughputs/throughputs_{instrument}_{mode}.pickle'
    with open(t_file, 'wb') as handle:
        pickle.dump(throughputs, handle, protocol=pickle.HIGHEST_PROTOCOL)


def generate_all_miri_throughputs():
    instrument = 'miri'
    mode = 'lrsslitless'
    readout = 'fastr1'
    disperser = 'p750'

    calc = jwst.Calculation(instrument, mode)
    calc.set_scene(
        sed_type='flat', sed_model='flam',
        norm_band='2mass,ks', norm_magnitude=15.0,
    )
    filters = calc.get_configs('filters')
    subarrays = calc.get_configs('subarrays')

    nint = 1
    ngroup = 2

    throughputs = {}
    for subarray in subarrays:
        filter = 'None'
        throughputs[subarray] = {}
        calc.perform_calculation(
            nint, ngroup, filter, readout, subarray, disperser,
        )
        wave, flux = calc.report['1d']['extracted_flux']
        wl_arr, qe = get_throughputs(calc, filter)

        norm = np.amax(flux) / np.amax(qe[filter])
        wl_mask, qe_mask = mask_througput(
            wl_arr, qe[filter], wave, flux/norm, thin=250,
        )
        throughputs[subarray][str(filter)] = {'wl': wl_mask, 'response': qe_mask}

    t_file = f'../data/throughputs/throughputs_{instrument}_{mode}.pickle'
    with open(t_file, 'wb') as handle:
        pickle.dump(throughputs, handle, protocol=pickle.HIGHEST_PROTOCOL)


def generate_all_niriss_throughputs():
    instrument = 'niriss'
    mode = 'soss'
    readout = 'nisrapid'

    calc = jwst.Calculation(instrument, mode)
    disperser = calc.get_configs('dispersers')[0]
    calc.set_scene(
        sed_type='flat', sed_model='flam',
        norm_band='2mass,ks', norm_magnitude=15.0,
    )
    filters = calc.get_configs('filters')
    subarrays = calc.get_configs('subarrays')

    nint = 1
    ngroup = 2

    # Could not generate the throughputs from pandeia
    # For the moment will use this digitized version from the jdocs
    wl, through = np.loadtxt(
        '../data/soss_GR700XD_throughput_etcv2.0_digitized.txt', unpack=True,
    )
    wl_o1 = wl[0:250]
    through_o1 = through[0:250]
    # Only for filter=clear, subarray=substrip256 v full
    wl_o2 = wl[250:]
    through_o2 = through[250:]

    fd = fits.getdata('../etc_files/wb_soss_o2/lineplot/lineplot_extracted_flux.fits')
    wave = fd['WAVELENGTH']
    flux = fd['extracted_flux']

    norm = np.amax(flux) / np.amax(through_o2)
    wl_mask_o2, qe_mask_o2 = mask_througput(
        wl_o2, through_o2, wave, flux/norm, thin=1,
    )

    plt.clf()
    plt.plot(wl_o2, through_o2, dashes=(8,2), lw=1.5, color='k')
    plt.plot(wl_mask, qe_mask, color='r', lw=1.75, alpha=0.75)


    throughputs = {}
    tmp_through = {}
    for subarray in subarrays:
        throughputs[subarray] = {}
        for filter in filters:
            print(subarray, disperser, filter)
            try:
                calc.perform_calculation(
                    nint, ngroup, filter, readout, subarray, disperser,
                )
            except:
                print(
                    f'Pandeia failed for: {instrument}, {mode}, '
                    f'{subarray}, {disperser}, {filter}'
                )
                continue

            wave, flux = calc.report['1d']['extracted_flux']
            #wl_arr, qe = get_throughputs(calc, [filter])
            tmp_through[filter] = {'wl': wave, 'flux': flux}
            if filter == 'clear':
                qe = np.copy(through_o1)
            elif filter == 'f277w':
                wl_min = np.amin(tmp_through['f277w']['wl'])
                wl_mask = tmp_through['clear']['wl']>= wl_min
                scale = (
                    tmp_through['f277w']['flux'] /
                    tmp_through['clear']['flux'][wl_mask]
                )
                interp_qe = sip.interp1d(
                    wave, scale, bounds_error=False, fill_value=0.0,
                )
                qe = through_o1 * interp_qe(wl_o1)

            norm = np.amax(flux) / np.amax(qe)
            wl_mask, qe_mask = mask_througput(
                wl_o1, qe, wave, flux/norm, thin=1,
            )
            throughputs[subarray][filter] = {'wl': wl_mask, 'response': qe_mask}

    # Add order 2
    filter = 'clear'
    for subarray in subarrays:
        if subarray == 'substrip96':
            continue
        print(subarray)
        throughputs[subarray][filter]['order2'] = {
            'wl': wl_mask_o2, 'response': qe_mask_o2,
        }

    t_file = f'../data/throughputs/throughputs_{instrument}_{mode}.pickle'
    with open(t_file, 'wb') as handle:
        pickle.dump(throughputs, handle, protocol=pickle.HIGHEST_PROTOCOL)


# compare

t_file = '../data/throughputs/throughputs_miri_lrsslitless.pickle'
t_file = '../data/throughputs/throughputs_nircam_ssgrism.pickle'
t_file = '../data/throughputs/throughputs_nirspec_bots.pickle'
t_file = '../data/throughputs/throughputs_niriss_soss.pickle'

with open(t_file, 'rb') as handle:
    data = pickle.load(handle)

subarrays = list(data.keys())
filters = list(data[subarrays[0]].keys())

filter = filters[0]
plt.figure(0)
plt.clf()
for i,subarray in enumerate(subarrays):
    for filter in filters:
        wl = data[subarray][filter]['wl']
        response = data[subarray][filter]['response'] * (1+i*0.03)
        plt.plot(wl, response, lw=1.5, label=subarray + ' / ' + filter)
#plt.xlim(0.9*np.amin(wl_mask), 1.1*np.amax(wl_mask))
plt.legend(loc='best')
plt.ylim(bottom=0.0)
plt.title(filter)
plt.xlabel('wavelength')
plt.ylabel('throughput')
plt.tight_layout()








# Copyright (c) 2024 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

__all__ = [
    'find_target',
    'Catalog',
    'load_trexolists',
    'load_targets',
    'load_aliases',
]

from datetime import datetime
from astropy.io import ascii
import numpy as np
import prompt_toolkit as ptk

from ..utils import ROOT
from . import utils as u
from .target import Target


def find_target(targets=None):
    """
    Interactive prompt with tab-completion to search for targets.

    Parameters
    ----------
    targets: list of Target objects

    Examples
    --------
    >>> import gen_tso.catalogs as cat
    >>> target = cat.find_target()
    """
    if targets is None:
        targets = load_targets('nea_data.txt', is_confirmed=True)
    planets = [target.planet for target in targets]
    aliases = []
    for target in targets:
        aliases += target.aliases
    planets += list(aliases)

    completer = ptk.completion.WordCompleter(
        planets,
        sentence=True,
        match_middle=True,
    )
    session = ptk.PromptSession(
        history=ptk.history.FileHistory(f'{ROOT}/data/target_search_history')
    )
    name = session.prompt(
        "(Press 'tab' for autocomplete)\nEnter Planet name: ",
        completer=completer,
        complete_while_typing=False,
    )
    if name in aliases:
        for target in targets:
            if name in target.aliases:
                return target
    if name in planets:
        return targets[planets.index(name)]

    return None


class Catalog():
    """
    Load the entire catalog.

    Examples
    --------
    >>> import gen_tso.catalogs as cat
    >>> catalog = cat.Catalog()
    """
    def __init__(self):
        # Confirmed planets and TESS candidates
        nea_targets = load_targets('nea_data.txt', is_confirmed=True)
        tess_targets = load_targets('tess_data.txt', is_confirmed=False)
        self.targets = nea_targets + tess_targets

        # JWST targets
        trexo_data = load_trexolists(grouped=True)
        njwst = len(trexo_data)
        host_aliases = load_aliases('host')

        jwst_hosts = []
        for jwst_target in trexo_data:
            hosts = np.unique([
                host_aliases[host] if host in host_aliases else host
                for host in jwst_target['target']
            ])
            jwst_target['nea_hosts'] = hosts
            jwst_hosts += list(hosts)
        jwst_hosts = np.unique(jwst_hosts)

        planet_aliases = load_aliases('planet')
        planets_aka = u.invert_aliases(planet_aliases)

        for target in self.targets:
            target.is_jwst = target.host in jwst_hosts and target.is_transiting
            if target.is_jwst:
                for j in range(njwst):
                    if target.host in trexo_data[j]['nea_hosts']:
                        break
                target.trexo_data = trexo_data[j]

            if target.planet in planets_aka:
                target.aliases = planets_aka[target.planet]

        self._transit_mask = [target.is_transiting for target in self.targets]
        self._jwst_mask = [target.is_jwst for target in self.targets]
        self._confirmed_mask = [target.is_confirmed for target in self.targets]


    def get_target(
            self, name=None,
            is_transit=True, is_jwst=None, is_confirmed=True,
        ):
        """
        Search by name for a planet in the catalog.

        Parameters
        ----------
        name: String
            If not None, name of the planet to search.
            If None, an interactive prompt will open to search for the planet
        is_transit: Bool
            If True/False restrict search to transiting/non-transiting planets
            If None, consider all targets.
        is_jwst: Bool
            If True/False restrict search to planet of/not JWST hosts
            If None, consider all targets.
        is_confirmed: Bool
            If True/False restrict search to confirmed/candidate planets
            If None, consider all targets.

        Returns
        -------
        target: a Target object
            Target with the system properties of the searched planet.
            If no target was found on the catalog, return None.
        """
        mask = np.ones(len(self.targets), bool)
        if is_transit is not None:
            mask &= np.array(self._transit_mask) == is_transit
        if is_jwst is not None:
            mask &= np.array(self._jwst_mask) == is_jwst
        if is_confirmed is not None:
            mask &= np.array(self._confirmed_mask) == is_confirmed

        targets = [target for target,flag in zip(self.targets,mask) if flag]

        if name is None:
            return find_target(targets)

        target = u.normalize_name(name)
        for target in targets:
            if name == target.planet or name in target.aliases:
                return target

    def show_target(
            self, name=None,
            is_transit=True, is_jwst=None, is_confirmed=True,
        ):
        target = self.get_target(name, is_transit, is_jwst, is_confirmed)
        if target is None:
            return
        print(target)


def load_targets(database='nea_data.txt', is_confirmed=np.nan):
    """
    Unpack star and planet properties from plain text file.

    Parameters
    ----------
    databases: String
        nea_data.txt or tess_data.txt

    Returns
    -------
    targets: List of Target

    Examples
    --------
    >>> import gen_tso.catalogs as cat
    >>> nea_data = cat.load_nea_targets_table()
    """
    # database = 'new_nea_data.txt'
    with open(f'{ROOT}data/{database}', 'r') as f:
        lines = f.readlines()

    lines = [
        line for line in lines
        if not line.strip().startswith('#')
    ]
    targets = []
    for line in lines:
        if line.startswith('>'):
            name_len = line.find(':')
            host = line[1:name_len]
            star_vals = np.array(line[name_len+1:].split(), float)
            ra, dec, ks_mag, rstar, mstar, teff, logg, metal = star_vals
        elif line.startswith(' '):
            name_len = line.find(':')
            planet = line[1:name_len].strip()
            planet_vals = np.array(line[name_len+1:].split(), float)
            t_dur, rplanet, mplanet, sma, period, teq, min_mass = planet_vals

            target = Target(
                host=host,
                mstar=mstar, rstar=rstar, teff=teff, logg_star=logg,
                metal_star=metal,
                ks_mag=ks_mag, ra=ra, dec=dec,
                planet=planet,
                mplanet=mplanet, rplanet=rplanet,
                period=period, sma=sma, transit_dur=t_dur,
                is_confirmed=is_confirmed,
                is_min_mass=bool(min_mass),
            )
            targets.append(target)

    return targets


def load_trexolists(grouped=False, trexo_file=None):
    """
    Get the data from the trexolists.csv file.
    Note that trexolists know targets by their host star, not by
    individual planets in a given system.

    Parameters
    ----------
    grouped: Bool
        - If False, return the a dictionary where each item contains
          a 1D list for each individual program.
        - If True, return list for each individual host, where each entry
          contains a dictionary of 1D lists of all programs for the target.
    trexo_file: String
        If None, extract data from default Gen TSO location.
        Otherwise, a path to a trexolists.csv file.

    Examples
    --------
    >>> import gen_tso.catalogs as cat
    >>>
    >>> # Get data as lists of individual programs:
    >>> trexo = cat.load_trexolists()
    >>> print(trexo['target'])
    ['L 168-9' 'HAT-P-14' 'WASP-80' 'WASP-80' 'WASP-69' 'GJ 436' ...]
    >>>
    >>> print(list(trexo))
    ['target', 'trexo_name', 'program', 'ra', 'dec', 'event', 'mode', 'subarray', 'readout', 'groups', 'phase_start', 'phase_end', 'duration', 'date_start', 'plan_window', 'proprietary_period', 'status']

    >>> # Get data as lists of (host) targets:
    >>> trexo = cat.load_trexolists(grouped=True)
    >>> trexo[19]
    {'target': array(['WASP-43', 'WASP-43'], dtype='<U23'),
    'trexo_name': array(['WASP-43', 'WASP-43'], dtype='<U23'),
    'program': array(['GTO 1224 Birkmann', 'ERS 1366 Batalha'], dtype='<U22'),
    'ra': array(['10:19:37.9634', '10:19:37.9649'], dtype='<U13'),
    'dec': array(['-09:48:23.21', '-09:48:23.19'], dtype='<U12'),
    'event': array(['phase', 'phase'], dtype='<U9'),
    'mode': array(['NIRSPEC BOTS+G395H', 'MIRI LRS'], dtype='<U20'),
    'subarray': array(['SUB2048', 'SLITLESSPRISM'], dtype='<U13'),
    'readout': array(['NRSRAPID', 'FASTR1'], dtype='<U8'),
    'groups': array([20, 64]),
    'phase_start': array([0.18394, 0.13912]),
    'phase_end': array([0.20955, 0.16473]),
    'duration': array([28.57, 31.82]),
    'date_start': array([datetime.datetime(2023, 5, 14, 9, 3, 30),
           datetime.datetime(2022, 11, 30, 23, 36, 1)], dtype=object),
    'plan_window': array(['X', 'X'], dtype='<U21'),
    'proprietary_period': array([12,  0]),
    'status': array(['Archived', 'Archived'], dtype='<U14'),
    'truncated_ra': array('10:19', dtype='<U5'),
    'truncated_dec': array('-09:48', dtype='<U6')}
    """
    if trexo_file is None:
        trexo_file = f'{ROOT}data/trexolists.csv'

    trexolist_data = ascii.read(
        trexo_file,
        format='csv', guess=False, fast_reader=False, comment='#',
    )

    norm_targets = np.array([
        u.normalize_name(target)
        for target in trexolist_data['Target']
    ])
    trexo_data = {
        'target': norm_targets,
        'trexo_name': np.array(trexolist_data['Target'])
    }

    category = trexolist_data['Category']
    programs = trexolist_data['Program']
    pi = trexolist_data['PI name']
    trexo_data['program'] = np.array([
        f"{categ} {prog} {name}"
        for categ,prog,name in zip(category, programs, pi)
    ])

    trexo_data['ra'] = np.array(trexolist_data['R.A. 2000'])
    trexo_data['dec'] = np.array(trexolist_data['Dec. 2000'])

    trexo_data['event'] = np.array([
        event.lower().replace('phasec', 'phase')
        for event in trexolist_data['Event']
    ])
    # see_phase?

    trexo_data['mode'] = np.array([
        obs.replace('.', ' ')
        for obs in trexolist_data['Mode']
    ])
    trexo_data['subarray'] = np.array(trexolist_data['Subarray'])
    trexo_data['readout'] = np.array(trexolist_data['Readout pattern'])
    trexo_data['groups'] = np.array(trexolist_data['Groups'])

    trexo_data['phase_start'] = np.array([
        np.nan if phase=='N/A' else float(phase)
        for phase in trexolist_data['Start.Phase']
    ])
    trexo_data['phase_end'] = np.array([
        np.nan if phase=='N/A' else float(phase)
        for phase in trexolist_data['End.Phase']
    ])
    trexo_data['duration'] = np.array(trexolist_data['Hours'])

    trexo_data['date_start'] = np.array([
        np.nan if date=='X' else datetime.strptime(date,'%b_%d_%Y_%H:%M:%S')
        for date in trexolist_data['Start date']
    ])

    date_start = []
    date_end = []
    plan_window = []
    s_dates = trexolist_data['Start date']
    e_dates = trexolist_data['End date']
    windows = trexolist_data['Plan Windows']
    for i in range(len(norm_targets)):
        try:
            date = datetime.strptime(s_dates[i],'%b_%d_%Y_%H:%M:%S')
        except:
            date = np.nan
        try:
            end = datetime.strptime(e_dates[i],'%b_%d_%Y_%H:%M:%S')
        except:
            end = np.nan
        try:
            window = windows[i][:windows[i].index('-')]
            window = datetime.strptime(window,'%b%d,%Y')
        except:
            window = np.nan
        date_start.append(date)
        date_end.append(end)
        plan_window.append(window)
    trexo_data['date_start'] = np.array(date_start)
    trexo_data['date_end'] = np.array(date_end)
    trexo_data['plan_window'] = np.array(plan_window)

    trexo_data['proprietary_period'] = np.array(trexolist_data['Prop.Period'])
    trexo_data['status'] = np.array(trexolist_data['Status'])

    if not grouped:
        return trexo_data

    # Use RA and dec to detect aliases for a same object
    truncated_ra = np.array([ra[0:5] for ra in trexo_data['ra']])
    truncated_dec = np.array([dec[0:6] for dec in trexo_data['dec']])

    ntargets = len(trexo_data['target'])
    taken = np.zeros(ntargets, bool)
    target_sets = []
    trexo_ra = []
    trexo_dec = []
    for i in range(ntargets):
        if taken[i]:
            continue
        group_indices = [i]
        ra = truncated_ra[i]
        dec = truncated_dec[i]
        trexo_ra.append(ra)
        trexo_dec.append(dec)
        taken[i] = True
        for j in range(i,ntargets):
            if truncated_ra[j]==ra and truncated_dec[j]==dec and not taken[j]:
                group_indices.append(j)
                taken[j] = True
        target_sets.append(group_indices)

    grouped_data = []
    for i,indices in enumerate(target_sets):
        target = {}
        for key in trexo_data.keys():
            target[key] = trexo_data[key][indices]
        target['truncated_ra'] = np.array(trexo_ra[i])
        target['truncated_dec'] = np.array(trexo_dec[i])
        grouped_data.append(target)

    return grouped_data


def parse(name, style):
    """
    Parse a planet name as a planet (no change) or as a star,
    in which case it will identify from confirmed or candidate
    format name.

    Parameters
    ----------
    style: String
        Select from 'planet' or 'host'.

    Returns
    -------
    name: String
        Parsed name.

    Examples
    --------
    >>> from gen_tso.catalogs.catalogs import parse
    >>> parse('WASP-80 b', 'planet')
    WASP-80 b'
    >>> parse('WASP-80 b', 'host')
    WASP-80'
    >>> parse('TOI-316.01', 'host')
    'TOI-316'
    """
    if style == 'planet':
        return name
    elif style == 'host':
        if u.is_letter(name):
            return name[:-2]
        end = name.rindex('.')
        return name[:end]


def load_aliases(style='planet', aliases_file=None):
    """
    Load file with known aliases of NEA targets.

    Parameters
    ----------
    style: String
        Select from 'planet', 'host', or 'system'.

    Returns
    -------
    aliases: Dictionary
        Dictionary of aliases to-from NASA Exoplanet Archive name.
        See below for examples depending on the 'style' argument.

    Examples
    --------
    >>> import gen_tso.catalogs as cat
    >>>
    >>> # From alias planet name to NEA name:
    >>> aliases = cat.load_aliases('planet')
    >>> aliases['CD-38 2551 b']
    'WASP-63 b'
    >>>
    >>> # From alias host name to NEA name:
    >>> aliases = cat.load_aliases('host')
    >>> aliases['CD-38 2551']
    'WASP-63'
    >>>
    >>> # As stellar system with all host and planet aliases:
    >>> aliases = cat.load_aliases('system')
    >>> aliases['WASP-63']
    {'host': 'WASP-63',
     'planets': ['WASP-63 b'],
     'host_aliases': array(['CD-38 2551', 'TOI-483', 'WASP-63'], dtype='<U10'),
     'planet_aliases': {'TOI-483.01': 'WASP-63 b',
      'CD-38 2551 b': 'WASP-63 b',
      'WASP-63 b': 'WASP-63 b'}}
    """
    if style not in ['planet', 'host', 'system']:
        raise ValueError(
            "Invalid alias style, select from: 'planet', 'host', or 'system'"
        )
    if aliases_file is None:
        aliases_file = f'{ROOT}data/target_aliases.txt'

    with open(aliases_file, 'r') as f:
        lines = f.readlines()

    if style != 'system':
        aliases = {}
        for line in lines:
            loc = line.index(':')
            name = parse(line[:loc], style)
            for alias in line[loc+1:].strip().split(','):
                aliases[parse(alias,style)] = name
            aliases[name] = name
        return aliases

    aliases = {}
    current_host = ''
    for line in lines:
        loc = line.index(':')
        planet = parse(line[:loc], 'planet')
        host = parse(line[:loc], 'host')
        host_aliases = [
            parse(name, 'host')
            for name in line[loc+1:].strip().split(',')
        ]
        host_aliases += [host]
        planet_aliases = {
            parse(name, 'planet'): planet
            for name in line[loc+1:].strip().split(',')
        }
        planet_aliases[planet] = planet
        if host != current_host:
            # Save old one
            if current_host != '':
                system['host_aliases'] = np.unique(system['host_aliases'])
                aliases[current_host] = system
            # Start new one
            system = {
                'host': host,
                'planets': [planet],
                'host_aliases': host_aliases,
                'planet_aliases': planet_aliases,
            }
            current_host = host
        else:
            system['host_aliases'] += host_aliases
            system['planets'] += [planet]
            system['planet_aliases'].update(planet_aliases)
    system['host_aliases'] = np.unique(system['host_aliases'])
    aliases[current_host] = system
    return aliases




# Copyright (c) 2024 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

__all__ = [
    'plotly_filters',
    'plotly_sed_spectra',
    'plotly_depth_spectra',
    'plotly_tso_spectra',
]

from itertools import groupby
import numpy as np
import plotly.express as px
import plotly.graph_objects as go
#import pyratbay.spectrum as ps
from pyratbay.tools import u
from .. import pandeia_io as jwst
from ..utils import (
    bin_spectrum,
    constant_resolution_spectrum,
)


COLOR_SEQUENCE = [
    'Royalblue', # blue
    '#15b01a',  # green
    '#000075',  # navy
    '#f032e6',  # magenta
    '#42d4f4',  # cyan
    '#888888',  # grey
    'Red',      # red
    '#9A6324',  # brown
    '#800000',  # maroon
    '#000000',  #  black
    '#469990',  # teal
    '#911eb4',  # purple
    '#808000',  # olive
    'Green',  # green
]


def band_boundaries(band, threshold=0.001):
    """
    Find the wavelength boundaries of a passband where the response
    is greater than the required threshold.

    Parameters
    ----------
    band: dict
        A dictionary of the band response and wavelength.
    threshold: float
        Minimum band response for highlight.

    Returns
    -------
    bounds: list of float pairs
        A list of the wavelength boundaries for each contiguous
        segment with non-zero response.
    """
    wl = band['wl']
    response = band['response']
    bounds = []
    # Contigous ranges where response > threshold:
    for group, indices in groupby(range(len(wl)), lambda x: response[x]>threshold):
        if group:
            indices = list(indices)
            imin = indices[0]
            imax = indices[-1]
            bounds.append((wl[imin], wl[imax]))
    return bounds


def plotly_filters(passbands, inst_name, mode, subarray, filter_name, show_all):
    """
    Make a plotly figure of the passband filters.
    """
    if inst_name is None:
        return go.Figure()

    instruments = [inst_name]
    if show_all:
        instruments += [
            inst for inst in passbands.keys()
            if inst not in instruments
        ]

    show_all_subarrays = {
        'miri': 'slitlessprism',
        'nircam': 'subgrism64',
        'nirspec': 'sub2048',
        'niriss': 'substrip256',
    }
    show_all_filters = {
        'miri': ['None'],
        'nircam': ['f322w2', 'f444w'],
        'nirspec': ['g140h/f100lp', 'g235h/f170lp', 'g395h/f290lp', 'prism/clear'],
        'niriss': ['clear'],
    }

    # Parse filters to plot
    all_filters = {}
    nfilters = 0
    for inst in instruments:
        all_filters[inst] = {}

        if inst != inst_name:
            subarray = show_all_subarrays[inst]
        if subarray not in passbands[inst]:
            return go.Figure()

        if inst == inst_name:
            filters = list(passbands[inst][subarray].keys())
        else:
            filters = show_all_filters[inst]

        for filter in filters:
            band = passbands[inst][subarray][filter]
            all_filters[inst][filter] = band
            if 'order2' in band:
                all_filters[inst]['order2'] = band['order2']
            nfilters += len(all_filters[inst][filter])


    visible = [None for _ in range(nfilters)]
    if mode == 'bots':
        for i,filter in enumerate(all_filters[inst_name].keys()):
            hide = ('h' in filter_name) is not ('h' in filter)
            if hide and 'prism' not in filter:
                visible[i] = 'legendonly'
    elif mode == 'ssgrism':
        for i,filter in enumerate(all_filters[inst_name].keys()):
            if filter != filter_name and filter not in ['f322w2', 'f444w']:
                visible[i] = 'legendonly'

    if mode == 'bots':
        primary_colors = px.colors.sample_colorscale(
            'Viridis', np.linspace(0, 0.8, 9),
        )
    else:
        primary_colors = px.colors.sample_colorscale(
            'Viridis', [0.1, 0.3, 0.5, 0.7],
        )
    secondary_colors = [
        px.colors.sequential.gray[3],
        px.colors.sequential.gray[5],
        px.colors.sequential.gray[7],
        px.colors.sequential.gray[8],
        px.colors.sequential.gray[9],
    ]
    secondary_nirspec = [
        px.colors.sequential.Magma[3],
        px.colors.sequential.Magma[5],
        px.colors.sequential.Magma[6],
        px.colors.sequential.Magma[7],
    ]
    sel_cols = iter(primary_colors)
    other_cols = iter(secondary_colors)
    nirspec_cols = iter(secondary_nirspec)

    fig = go.Figure()
    j = 0
    filter_index = -1
    for inst in instruments:
        for filter, throughput in all_filters[inst].items():
            if filter == filter_name:
                color = 'Gold'
                width = 3.5
                filter_index = j
            elif filter == 'order2' and filter_name == 'clear':
                color = 'Orange'
                width = 3.5
            elif inst == inst_name:
                color = next(sel_cols)
                width = 2.0
            else:
                color = next(nirspec_cols) if inst=='nirspec' else next(other_cols)
                width = 1.25
            linedict = dict(color=color, width=width)
            fig.add_trace(go.Scatter(
                x=throughput['wl'],
                y=throughput['response'],
                mode='lines',
                name=filter.upper(),
                legendgroup=inst,
                legendgrouptitle_text=inst,
                line=linedict,
                legendrank=j+nfilters*int(inst != inst_name),
                visible=visible[j],
            ))
            j += 1
    fig.update_traces(
        hovertemplate=
            'wl = %{x:.2f}<br>'+
            'throughput = %{y:.3f}'
    )
    fig.update_yaxes(
        title_text=None,
        autorangeoptions=dict(minallowed=0),
    )

    wl_scale = 'log'
    wl_range = [np.log10(0.6), np.log10(13.5)]
    wl_range = [np.log10(0.6), np.log10(28.0)]
    fig.update_xaxes(
        title_text='wavelength (um)',
        title_standoff=0,
        range=wl_range,
        type=wl_scale,
    )
    fig.update_layout(
        showlegend=True,
        legend=dict(groupclick="toggleitem"),
    )

    # Show current filter on top (TBD: there must be a better way)
    if filter_index < 0:
        return fig
    fig_idx = np.arange(len(fig.data))
    fig_idx[-1] = filter_index
    fig_idx[filter_index] = len(fig.data) - 1
    fig.data = tuple(np.array(fig.data)[fig_idx])
    return fig


def plotly_sed_spectra(
        sed_models, labels, highlight_model=None,
        wl_range=[0.5,28], units='mJy', wl_scale='linear', resolution=250.0,
        throughput=None,
    ):
    """
    Make a plotly figure of stellar SED spectra.
    """
    nmodels = len(sed_models)
    fig = go.Figure(
        #layout={'colorway':px.colors.qualitative.Alphabet},
        layout={'colorway':COLOR_SEQUENCE},
    )

    # Shaded area for filter:
    if throughput is not None:
        if 'order2' in throughput:
            band_bounds = band_boundaries(throughput['order2'], threshold=0.03)
            for bound in band_bounds:
                fig.add_vrect(
                    fillcolor="LightSalmon", opacity=0.4,
                    x0=bound[0], x1=bound[1],
                    layer="below", line_width=0,
                )
        band_bounds = band_boundaries(throughput)
        for bound in band_bounds:
            fig.add_vrect(
                fillcolor="#069af3", opacity=0.4,
                x0=bound[0], x1=bound[1],
                layer="below", line_width=0,
            )

    for j,model in enumerate(sed_models):
        wl = model['wl']
        flux = model['flux']
        if resolution > 0:
            wl_min = np.amin(wl)
            wl_max = np.amax(wl)
            bin_wl = constant_resolution_spectrum(wl_min, wl_max, resolution)
            bin_flux = bin_spectrum(bin_wl, wl, flux, ignore_gaps=True)
            mask = np.isfinite(bin_flux)
            wl = bin_wl[mask]
            flux = bin_flux[mask]

        if highlight_model is None:
            linedict = dict(width=1.25)
            rank = j
        elif labels[j] == highlight_model:
            linedict = dict(color='Gold', width=2.0)
            rank = j + nmodels
        else:
            linedict = dict(width=1.25)
            rank = j
        fig.add_trace(go.Scatter(
            x=wl,
            y=flux,
            mode='lines',
            opacity=0.75,
            name=labels[j],
            line=linedict,
            legendrank=rank,
        ))

    fig.update_traces(
        hovertemplate=
            'wl = %{x:.2f}<br>'+
            'flux = %{y:.3f}'
    )
    fig.update_yaxes(
        title_text=f'Flux ({units})',
        title_standoff=0,
    )

    if wl_scale == 'log':
        wl_range = [np.log10(wave) for wave in wl_range]
    fig.update_xaxes(
        title_text='wavelength (um)',
        title_standoff=0,
        range=wl_range,
        type=wl_scale,
    )

    fig.update_layout(legend=dict(
        orientation="h",
        entrywidth=1.0,
        entrywidthmode='fraction',
        yanchor="bottom",
        xanchor="right",
        y=1.02,
        x=1
    ))
    fig.update_layout(showlegend=True)
    return fig


def plotly_depth_spectra(
        depth_models, labels, highlight_model=None,
        wl_range=[0.5,12], units='percent', wl_scale='linear', resolution=250.0,
        obs_geometry='Transit',
        throughput=None,
    ):
    """
    Make a plotly figure of transit/eclipse depth spectra.
    """
    nmodels = len(depth_models)
    fig = go.Figure(
        layout={'colorway':COLOR_SEQUENCE},
    )
    # Shaded area for filter:
    if throughput is not None:
        if 'order2' in throughput:
            band_bounds = band_boundaries(throughput['order2'], threshold=0.03)
            for bound in band_bounds:
                fig.add_vrect(
                    fillcolor="LightSalmon", opacity=0.4,
                    x0=bound[0], x1=bound[1],
                    layer="below", line_width=0,
                )
        band_bounds = band_boundaries(throughput)
        for bound in band_bounds:
            fig.add_vrect(
                fillcolor="#069af3", opacity=0.4,
                x0=bound[0], x1=bound[1],
                layer="below", line_width=0,
            )

    for j,model in enumerate(depth_models):
        wl = model['wl']
        depth = model['depth']
        if resolution > 0:
            wl_min = np.amin(wl)
            wl_max = np.amax(wl)
            bin_wl = constant_resolution_spectrum(wl_min, wl_max, resolution)
            depth = bin_spectrum(bin_wl, wl, depth, ignore_gaps=True) / u(units)
            mask = np.isfinite(depth)
            wl = bin_wl[mask]
            depth = depth[mask]

        if labels[j] == highlight_model:
            linedict = dict(color='Gold', width=2.0)
            rank = j + nmodels
        else:
            linedict = dict(width=1.25)
            rank = j
        fig.add_trace(go.Scatter(
            x=wl,
            y=depth,
            mode='lines',
            name=labels[j],
            line=linedict,
            legendrank=rank,
        ))

    fig.update_traces(
        hovertemplate=
            'wl = %{x:.2f}<br>'+
            'depth = %{y:.3f}'
    )
    fig.update_yaxes(
        title_text=f'{obs_geometry} depth ({units})'.replace('percent','%'),
        title_standoff=0,
    )

    if wl_scale == 'log':
        wl_range = [np.log10(wave) for wave in wl_range]
    fig.update_xaxes(
        title_text='wavelength (um)',
        title_standoff=0,
        range=wl_range,
        type=wl_scale,
    )

    fig.update_layout(legend=dict(
        orientation="h",
        entrywidth=1.0,
        entrywidthmode='fraction',
        yanchor="bottom",
        xanchor="right",
        y=1.02,
        x=1
    ))
    fig.update_layout(showlegend=True)
    return fig


def plotly_tso_spectra(
        tso_list, resolution, n_obs, model_label, instrument_label,
        bin_widths=None,
        units='percent', wl_range=None, wl_scale='linear',
        obs_geometry='Transit',
    ):
    """
    Make a plotly figure of transit/eclipse depth TSO spectra.
    """
    if not isinstance(tso_list, list):
        tso_list = [tso_list]

    if isinstance(instrument_label, str):
        instrument_label = [instrument_label for tso in tso_list]

    fig = go.Figure()
    obs_col = px.colors.sample_colorscale('Viridis', 0.2)[0]
    model_col = px.colors.sample_colorscale('Viridis', 0.75)[0]

    ymax = 0.0
    ymin = np.inf
    legends = []
    for i,tso in enumerate(tso_list):
        bin_wl, bin_spec, bin_err, widths = jwst.simulate_tso(
           tso, n_obs=n_obs, resolution=resolution, noiseless=False,
        )
        wl = tso['wl']
        spec = tso['depth_spectrum']

        show_legend = model_label not in legends
        fig.add_trace(go.Scatter(
            x=wl,
            y=spec/u(units),
            mode='lines',
            name=model_label,
            legendgroup=model_label,
            showlegend=show_legend,
            line=dict(color=model_col, width=1.5),
        ))
        legends.append(model_label)

        show_legend = instrument_label[i] not in legends
        fig.add_trace(go.Scatter(
            x=bin_wl,
            y=bin_spec/u(units),
            error_y=dict(type='data', array=bin_err/u(units), visible=True),
            mode='markers',
            name=instrument_label[i],
            legendgroup=instrument_label[i],
            showlegend=show_legend,
            marker=dict(color=obs_col, size=5),
        ))
        legends.append(instrument_label[i])
        ymax = np.amax([ymax, np.amax(spec)])
        ymin = np.amin([ymin, np.amin(spec)])

    fig.update_traces(
        hovertemplate=
            'wl = %{x:.2f}<br>'+
            'depth = %{y:.3f}'
    )
    ymax = ymax/u(units)
    ymin = ymin/u(units)
    dy = 0.1 * (ymax-ymin)
    y_range = [ymin-dy, ymax+dy]
    title = f'{obs_geometry} depth ({units})'
    title = title.replace('percent','%').replace(' (none)', '')
    fig.update_yaxes(
        title_text=title,
        title_standoff=0,
        range=y_range,
    )

    if wl_scale == 'log' and wl_range is not None:
        wl_range = [np.log10(wave) for wave in wl_range]
    fig.update_xaxes(
        title_text='wavelength (um)',
        title_standoff=0,
        range=wl_range,
        type=wl_scale,
    )

    fig.update_layout(legend=dict(
        orientation="h",
        entrywidth=1.0,
        entrywidthmode='fraction',
        yanchor="bottom",
        xanchor="right",
        y=1.02,
        x=1
    ))
    fig.update_layout(showlegend=True)
    return fig


